#ifndef __PL011_H__
#define __PL011_H__

#include <type.h>
#include <gpio.h>

/*
 * PL011:
 * · Separate 18x8 tx, 16x12 rx FIFO.
 * · Programable baud rate (UARTCLK/16).
 * · Start, stop & parity bits (added prior tx & removed after rx).
 * · False bit detection.
 * · Line break generation & detection.
 * · Support CTS & RTS
 * · Flow control
 * · Data 5, 6, 7 or 8 bits
 * · Even, odd, stick, no-parity
 * · 1, 2 stop-bits
 * 
 */

typedef struct PL011 PL011;

typedef union PL011_DR PL011_DR;
typedef union PL011_RSRECR PL011_RSRECR;
typedef union PL011_FR PL011_FR;
typedef union PL011_IBRD PL011_IBRD;
typedef union PL011_FBRD PL011_FBRD;
typedef union PL011_LCRH PL011_LCRH;
typedef union PL011_CR PL011_CR;
typedef union PL011_IFLS PL011_IFLS;

typedef union PL011_ICR PL011_ICR;

/* Data Register */
union PL011_DR
{
	struct
	{
		word_t data:8;
		word_t framing_error:1;
		word_t parity_error:1;
		word_t break_error:1;
		word_t overrun_error:1;
		word_t padding:20;
		
	} fields;
	
	word_t bits;
};

/* Receive Status Register/Error Clear Register */
union PL011_RSRECR
{
	struct
	{
		word_t framing_error:1;
		word_t parity_error:1;
		word_t break_error:1;
		word_t overrun_error:1;
		word_t padding:28;
		
	} fields;
	
	word_t bits;
};

/* Flag Register */
union PL011_FR
{
	struct
	{
		word_t cts:1;
		word_t dsr:1;
		word_t dcd:1;
		word_t busy:1;
		word_t rx_fifo_empty:1;
		word_t tx_fifo_full:1;
		word_t rx_fifo_full:1;
		word_t tx_fifo_empty:1;
		word_t ri:1;
		word_t padding:23;
		
	} fields;
	
	word_t bits;
};

/* Disabled IrDA Register */
typedef word_t PL011_ILPR;

/* Integer Baud Rate Divisor */
union PL011_IBRD
{
	struct
	{
		half_t idiv;
		half_t padding;
		
	} fields;
	
	word_t bits;
};

/* Fractional Baud Rate Divisor */
union PL011_FBRD
{
	struct
	{
		word_t fdiv:6;
		word_t padding:26;
		
	} fields;
	
	word_t bits;
};

/* Line Control Register */
union PL011_LCRH
{
	struct
	{
		word_t send_break:1;
		word_t parity_enable:1;
		word_t parity_select:1;
		word_t stop_bits:1;
		word_t fifo_enable:1;
		word_t word_len:2;
		word_t stick_parity_select:1;
		word_t padding:24;
		
	} fields;
	
	word_t bits;
};

/* Control Register */

// 0x301 = 0b0 0 11 0 0000 0 0 1
union PL011_CR
{
	struct
	{
		word_t enable:1;
		word_t siren:1;
		word_t sirlp:1;
		word_t padding_one:4;
		word_t loopback_enable:1;
		word_t tx_enable:1;
		word_t rx_enable:1;
		word_t dtr:1;
		word_t rts:1;
		word_t out_1:1;
		word_t out_2:1;
		word_t rts_enable:1;
		word_t cts_enable:1;
		word_t padding_two:16;
		
	} fields;
	
	word_t bits;
};

/* Interrupt FIFO level select */

#define FIFO_TX_1div8 0b000
#define FIFO_TX_1div4 0b001
#define FIFO_TX_1div2 0b010
#define FIFO_TX_3div4 0b011
#define FIFO_TX_7div8 0b100

#define FIFO_RX_1div8 0b000
#define FIFO_RX_1div4 0b001
#define FIFO_RX_1div2 0b010
#define FIFO_RX_3div4 0b011
#define FIFO_RX_7div8 0b100

union PL011_IFLS
{
	struct
	{
		word_t tx_fifo_lvl:3;
		word_t rx_fifo_lvl:3;
		word_t txifpsel:3;
		word_t rxifpsel:3;
		word_t padding:20;
		
	} fields;
	
	word_t bits;
};

/* Interrupt Clear */
union PL011_ICR
{
	struct
	{
		word_t ri:1;
		word_t cts:1;
		word_t dcd:1;
		word_t dsr:1;
		word_t rx:1;
		word_t tx:1;
		word_t rx_timeout:1;
		word_t framing_error:1;
		word_t parity_error:1;
		word_t break_error:1;
		word_t overrun_error:1;
		word_t padding:21;
		
	} fields;
	
	word_t bits;
};

#define PL011_ADDR  0x20201000

struct PL011
{
	PL011_DR 	dr;
	PL011_RSRECR rsrecr;
	
	byte_t 		padding_one[0x10];
	
	PL011_FR 	fr;
	
	byte_t 		padding_two[0x04];
	
	word_t 		ilpr; // not in use
	PL011_IBRD 	ibrd;
	PL011_FBRD 	fbrd;
	PL011_LCRH 	lcrh;
	PL011_CR 	cr;
	PL011_IFLS 	ifls;
	word_t 		imsc;
	word_t 		ris;
	word_t 		mis;
	PL011_ICR 	icr;
	word_t 		dmacr;
	
	byte_t 		padding_three[0x34];
	
	word_t 		itcr;
	word_t 		itip;
	word_t 		itop;
	word_t 		tdr;
};


void PL011_init();
void PL011_putch(char ch);
char PL011_getch();

#endif /* __PL011_H__ */
