#include <arm.h>

/**/

inline void irqEnable()
{
	__asm__("cpsie i");
}

inline void fiqEnable()
{
	__asm__("cpsie f");
}

inline void irqDisable()
{
	__asm__("cpsid i");
}

inline void fiqDisable()
{
	__asm__("cpsid f");
}

/**/

inline void mmu_enable()
{
	uint32_t ctrl;
		
	icache_disable();
	
	ctrl = getControlRegister();
	ctrl |= 0x00000001; 	// MMU enabled
	ctrl &= ~0x00000004; 	// Data cache disabled
	ctrl &= ~0x00001000; 	// Instruction cache disabled
	ctrl &= ~0x00000008; 	// Write buffer disabled
	ctrl |= 0x00000100; 	// MMU protection enabled
	ctrl |= 0x00800000; 	// Subpage AP bits disabled (ARMv6 mode)
	setControlRegister(ctrl);
}

inline void mmu_disable()
{
	// TODO: leer procedimento
	// uint32_t ctrl = getControlRegister();
	// ctrl &= ~0x00000001; // MMU enabled
	// ctrl &= ~0x00000004; // Level 1 data cache enabled
	// ctrl &= ~0x00000008; // Write buffer enabled
	// ctrl &= ~0x00000100; // MMU protection enabled
	// ctrl &= ~0x00001000; // Instruction cache enabled
	// ctrl &= ~0x00800000; // Subpage AP bits disabled (ARMv6 mode)
	// setControlRegister(ctrl);
}

inline void icache_disable()
{
	uint ctrl;
	ctrl = getControlRegister();
	ctrl &= ~0x00001000; // instruction cache disabled
	setControlRegister(ctrl);

	icache_invalidate();
}

inline void icache_invalidate()
{
	uint Rd = 0;
	__asm__ ("mcr p15, 0, %[Rd], c7, c5, 0" :: [Rd]"r"(Rd));
}

inline void tlb_flush()
{
	uint32_t Rd = 0;
	__asm__("mcr p15, 0, %[Rd], c8, c7, 0" :: [Rd]"r"(Rd));
}

inline void tlb_line_flush(uint32_t line)
{
	__asm__ ("mcr p15, 0, %[Rd], c8, c7, 1" :: [Rd]"r"(line));
}

/**/

inline void setControlRegister(uint32_t value)
{
	__asm__ ("mcr p15, 0, %[value], c1, c0, 0" :: [value] "r" (value));
}

inline uint32_t getControlRegister()
{
	uint32_t value = 0;
	__asm__ ("mrc p15, 0, %[value], c1, c0, 0" : [value]"=r"(value));
	return value;
}

/**/

inline void setTranslationTableBase0(uint32_t addr)
{
	addr &= ~0x3fff;
	__asm__("mcr p15, 0, %[addr], c2, c0, 0" :: [addr]"r"(addr));
}

inline void setTranslationTableBase1(uint32_t addr)
{
	addr &= ~0x3fff;
	__asm__("mcr p15, 0, %[addr], c2, c0, 1" :: [addr]"r"(addr));
}

inline void setTranslationTableControl(uint32_t value)
{
	__asm__("mcr p15, 0, %[value], c2, c0, 2" :: [value]"r"(value));
}

inline void setDomainAccessControl(uint32_t value)
{
	__asm__("mcr p15, 0, %[value], c3, c0, 0" :: [value]"r"(value));
}

inline uint32_t getTranslationTableBase0()
{
	uint32_t addr = 0;
	__asm__("mrc p15, 0, %[addr], c2, c0, 0" : [addr]"=r"(addr));
	return addr & ~0x3fff;
}

inline uint32_t getTranslationTableBase1()
{
	uint32_t addr = 0;
	__asm__("mrc p15, 0, %[addr], c2, c0, 1" : [addr]"=r"(addr));
	return addr & ~0x3fff;
}

inline uint32_t getTranslationTableControl()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c2, c0, 2" : [value]"=r"(value));
	return value;
}

inline uint32_t getDomainAccessControl()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c3, c0, 0" : [value]"=r"(value));
	return value;
}

/**/

inline uint32_t getDataFaultStatus()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c5, c0, 0" : [value]"=r"(value));
	return value;
}

inline uint32_t getInstructionFaultStatus()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c5, c0, 1" : [value]"=r"(value));
	return value;
}

inline uint32_t getFaultAddress()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c6, c0, 0" : [value]"=r"(value));
	return value;
}

inline uint32_t getInstructionFaultAddress()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c6, c0, 2" : [value]"=r"(value));
	return value;
}

/**/

inline uint32_t getInterruptStatus()
{
	uint32_t value = 0;
	__asm__("mrc p15, 0, %[value], c12, c1, 0" : [value]"=r"(value));
	return value;
}

/**/

inline void dataMemoryBarrier()
{
	uint32_t Rd = 0;
	__asm__("mcr p15, 0, %[Rd], c7, c10, 5" :: [Rd]"r"(Rd));
}

/* Debug */

const char *controlRegisterText[] = {
	"Memory Management Unit",
	"Alignment Fault Checking",
	"Write Buffer",
	"-",
	"-",
	"-",
	"-",
	"Endianness",
	"System Protection",
	"ROM Protection",
	"-",
	"Program Flow Prediction",
	"Level 1 Instruction Cache",
	"Exceptions Vector Location",
	"Replacement Strategy for Instruction and Data Cache",
	"T Bit",
	"Global Data TCM",
	"Reserved",
	"Global Instruction TCM",
	"-",
	"-",
	"Fast Interrupt Configuration",
	"Unaligned Data Access Operations",
	"Extended Page Table Configuration",
	"Configure Vectored Interrupt",
	"-",
	"-",
	"TEX Remap",
	"Access Flag"
};

void debugControlRegister(uint32_t ctrl)
{
	debug("[db] Control Register\n\r");
	debug("\t%s %s\n\r", controlRegisterText[0], (ctrl & (1<<0))? "enabled" : "disabled");
	debug("\t%s %s\n\r", controlRegisterText[1], (ctrl & (1<<1))? "enabled" : "disabled");
	debug("\t%s %s\n\r", controlRegisterText[2], (ctrl & (1<<2))? "enabled" : "disabled");
	debug("\t%s: %s\n\r", controlRegisterText[7], (ctrl & (1<<7))? "Big-endian" : "Little-endian");
	debug("\t%s %s\n\r", controlRegisterText[8], (ctrl & (1<<8))? "enabled" : "disabled");
	debug("\t%s %s\n\r", controlRegisterText[9], (ctrl & (1<<9))? "enabled" : "disabled");
	debug("\t%s %s\n\r", controlRegisterText[11], (ctrl & (1<<11))? "enabled" : "disabled");
	debug("\t%s %s\n\r", controlRegisterText[12], (ctrl & (1<<12))? "enabled" : "disabled");
	debug("\t%s: %s\n\r", controlRegisterText[13], (ctrl & (1<<13))? "high" : "normal");
	debug("\t%s: %s\n\r", controlRegisterText[14], (ctrl & (1<<14))? "Round-Robing" : "Random");
	debug("\t%s: %s\n\r", controlRegisterText[15], (ctrl & (1<<15))? "Set T bit" : "Don't set T bit");
	debug("\t%s: %s\n\r", controlRegisterText[21], (ctrl & (1<<21))? "Low interrupt latency" : "All performance features");
	debug("\t%s %s\n\r", controlRegisterText[22], (ctrl & (1<<22))? "enabled" : "disabled");
	debug("\t%s: %s\n\r", controlRegisterText[23], (ctrl & (1<<23))? "AP bits disabled" : "AP bits enabled");
	debug("\t%s: %s\n\r", controlRegisterText[24], (ctrl & (1<<24))? "Defined by VIC interface" : "Fixed");
	debug("\t%s %s\n\r", controlRegisterText[28], (ctrl & (1<<28))? "enabled" : "disabled");

}

