#include <heap.h>

Heap* heapInit(uint32_t ttb, uint32_t vAddress)
{
	uint attr;

	// asignar attr segun sistema/proceso
	if(vAddress == KERNEL_HEAP)
		attr = PAGE_RW_NA | PAGE_NOT_EXEC | PAGE_WR_THROUGH;
	else
		attr = PAGE_RW_RW | PAGE_NOT_EXEC | PAGE_WR_THROUGH;

	uint32_t frame = mmap_alloc_user();

	mmu_map_smallpage(ttb, frame, vAddress, attr);
	mmu_map_smallpage(getTranslationTableBase0(), frame, AUX_FRAME, attr);
		
	HeapNode *node = (HeapNode *) (AUX_FRAME + sizeof(Heap));
	
	node->size = SMALL_PAGE_SIZE - sizeof(HeapNode) - sizeof(Heap);
	node->prev = 0;
	node->next = 0;
	
	Heap *heap = (Heap *) AUX_FRAME;
	
	heap->freeByAddress = (HeapNode *) (vAddress + sizeof(Heap));
	heap->used = 0;
	heap->start = vAddress;
	heap->end = vAddress + SMALL_PAGE_SIZE;
        
	mmu_unmap_smallpage(getTranslationTableBase0(), AUX_FRAME);
	
	return (Heap*) vAddress;
}

void heapDeleteFree(Heap *heap, HeapNode *node)
{
	HeapNode *prev = 0, *next = 0;
	
	prev = node->prev;
	next = node->next;
	
	/* No es el primero */
	if(prev != 0)
	{
		/* No es el último */
		if(next != 0)
		{
			next->prev = prev;
			prev->next = next;
		}
		/* Es el último */
		else
		{
			prev->next = 0;
		}
	}
	/* Es el primero */
	else
	{
		/* No es el último */
		if(next != 0)
		{
			heap->freeByAddress = next;
			next->prev = 0;
		}
		/* Es el último */
		else
		{
			heap->freeByAddress = 0;
		}
	}
	
	// poner en 0 los links
	node->prev = 0;
	node->next = 0;
}

/* Inserción por dirección (menor a mayor) */
void heapInsertFree(Heap *heap, HeapNode *node)
{
	HeapNode *oldCursor = 0;
	HeapNode *cursor = heap->freeByAddress;
	
	/* Hay al menos un nodo */
	if(cursor != 0)
	{
		do
		{
			/* Si la dirección de nodo es > que la del nodo actual, saltear nodo */
			if(node > cursor)
			{
				oldCursor = cursor; // Me guardo una copia del anterior por si sale del bucle
				cursor = cursor->next;
				continue;
			}
			/* La dirección del nodo cursor es mayor insertar adelante de él */
			else
			{
				HeapNode *prev;
				
				prev = cursor->prev;
				
				/* Hay un nodo delante */
				if(prev != 0)
				{
					prev->next = node;
					node->prev = prev;
					node->next = cursor;
					cursor->prev = node;
				}
				/* No hay nodo delante, es el primero de la lista */
				else
				{
					heap->freeByAddress = node;
					node->next = cursor;
					cursor->prev = node;
				}
				
				/* Inserción completa */
				return;
			}
			
			
		} while(cursor != 0);
		
		/* salió por cursor == 0 ==> es el último de la lista, meter atrás de oldCursor */
		
		oldCursor->next = node;
		node->prev = oldCursor;
		
	}
	/* No hay nodos */
	else
	{
		heap->freeByAddress = node;
	}
}

void heapDeleteUsed(Heap *heap, HeapNode *node)
{
	HeapNode *prev = 0, *next = 0;
	
	prev = node->prev;
	next = node->next;
	
	/* No es el primero */
	if(prev != 0)
	{
		/* No es el último */
		if(next != 0)
		{
			next->prev = prev;
			prev->next = next;
		}
		/* Es el último */
		else
		{
			prev->next = 0;
		}
	}
	/* Es el primero */
	else
	{
		/* No es el último */
		if(next != 0)
		{
			heap->used = next;
			next->prev = 0;
		}
		/* Es el último */
		else
		{
			heap->used = 0;
		}
	}
	
	// poner en 0 los links
	node->prev = 0;
	node->next = 0;
}

/* Añadir adelante */
void heapInsertUsed(Heap *heap, HeapNode *node)
{
	HeapNode *cursor = heap->used;
	
	/* Si hay nodos */
	if(cursor != 0)
	{
		heap->used = node;
		node->next = cursor;
		cursor->prev = node;
	}
	/* Si no hay */
	else
	{
		heap->used = node;
	}
}

/* Devuelve el nodo remanente */
HeapNode* heapSplitNode(Heap *heap, HeapNode *node, uint32_t size)
{
	HeapNode *free;
	
	free = (HeapNode *) ((uint32_t) node + sizeof(HeapNode) + size);
	
	free->size = node->size - size - sizeof(HeapNode);
	free->prev = 0;
	free->next = 0;
	
	node->size = size;
	
	return free;
}

HeapNode* worstFitPolicy(Heap *heap)
{
	HeapNode *cursor = heap->freeByAddress;
	HeapNode *worstFit = cursor;
	
	while(cursor->next != 0)
	{
		cursor = cursor->next;

		if(cursor->size > worstFit->size)
			worstFit = cursor;
	}
	
	return worstFit;
}

int heapGrow(Heap *heap)
{
	// /** Si es el heap de sistema */
	uint attr;

	if((uint) heap == KERNEL_HEAP)
		attr = PAGE_RW_NA | PAGE_NOT_EXEC | PAGE_WR_THROUGH;
	else
		attr = PAGE_RW_RW | PAGE_NOT_EXEC | PAGE_WR_THROUGH;

	mmu_map_smallpage(getTranslationTableBase0(), mmap_alloc_user(), heap->end, attr);

	HeapNode *node = (HeapNode *) heap->end;

	heap->end += SMALL_PAGE_SIZE;

	node->size = SMALL_PAGE_SIZE - sizeof(HeapNode);
	node->prev = 0;
	node->next = 0;

	heapInsertFree(heap, node);
	heapMergeNode(heap, node);

	// debug("heapGrow()\n");
	// debugHeap(heap);
	return 0;
}

void* heapAllocateMemory(Heap *heap, uint32_t size)
{
	int iter = MaxIterations;
	HeapNode *node = 0;

	size = ((size & ~0x3) != size)? ((size & ~0x3) + 4) : size;
	
	do
	{
		node = worstFitPolicy(heap);
		
		if(node->size == size)
		{
			/* Utilizar nodo */
			heapDeleteFree(heap, node);
			heapInsertUsed(heap, node);
			//debugHeap(heap);
			return (void *) ((uint32_t) node + sizeof(HeapNode));
		}
		else if(node->size >= size + sizeof(HeapNode))
		{
			/* Partir nodo */
			heapDeleteFree(heap, node);
			HeapNode *free = heapSplitNode(heap, node, size);
			heapInsertFree(heap, free);
			heapInsertUsed(heap, node);
			//debugHeap(heap);
			return (void *) ((uint32_t) node + sizeof(HeapNode));
		}
		else
		{
			/* Agrandar heap */
			if(heapGrow(heap) < 0)
			{
				// debug("[db] heapGrow(): falló\n\r");
				//debugHeap(heap);
				return (void *) 0;
			}
		}
		
	} while(--iter);
	
	debug("[db] heapAllocateMemory(): salió por iteraciones\n\r");
	//debugHeap(heap);
	return (void *) 0;
}

void heapMergeNode(Heap *heap, HeapNode *node)
{
	HeapNode *prev, *next;
	
	prev = node->prev;
	next = node->next;
	
	/* Si hay anterior */
	if(prev != 0)
	{
		/* Verificar si son contiguos */
		if(((uint32_t) prev + sizeof(HeapNode) + prev->size) == (uint32_t) node)
		{
			/* Unir prev y node */
			prev->size += sizeof(HeapNode) + node->size;
			prev->next = next;
			
			if(next != 0)
				next->prev = prev;
			
			// node pasa a ser prev
			node = prev;
		}
	}
	
	// En este punto, o node se unio a prev o quedo igual
	
	if(next != 0)
	{
		if(((uint32_t) node + sizeof(HeapNode) + node->size) == (uint32_t) next)
		{
			node->size += sizeof(HeapNode) + next->size;
			node->next = next->next;
			
			if(next->next != 0)
				((HeapNode *) next->next)->prev = node;
		}
	}
}

void heapFreeMemory(Heap *heap, void *address)
{
	HeapNode *node = (HeapNode *) ((uint32_t) address - sizeof(HeapNode));
	heapDeleteUsed(heap, node);
	heapInsertFree(heap, node);
	heapMergeNode(heap, node);
}

void heapFreeAll(Heap *heap)
{
    HeapNode *node = heap->used;
    
    while(node != 0)
    {
        heapDeleteUsed(heap, node);
	heapInsertFree(heap, node);
	heapMergeNode(heap, node);
        node = heap->used;
    }
}

/* Debug */

void debugHeap(Heap *heap)
{
	HeapNode *free = heap->freeByAddress;
	
	debug("[db] Free list by ascending address\n\r");
	
	while(free != 0)
	{
		debug("\t %x - { .size = %d, .prev = %x, .next = %x };\n\r", (uint32_t) free, free->size, (uint32_t) free->prev, (uint32_t) free->next);
		free = free->next;
	}
	
	HeapNode *used = heap->used;
	
	debug("[db] Used list\n\r");
	
	while(used != 0)
	{
		debug("\t %x - { .size = %d, .prev = %x, .next = %x };\n\r", (uint32_t) used, used->size, (uint32_t) used->prev, (uint32_t) used->next);
		used = used->next;
	}
}

