#include <mmu.h>

void mmu_init()
{
	uint ttb, addr, doms;

	/* translation table */
	ttb = mmap_alloc_ttb();
	for(int i = 0; i < 4096; i++) ((uint *) ttb)[i] = FAULT;

	/* kernel */
	addr = 0;
	for(int i = 0; i < 4; i++)
	{
		mmu_map_section(ttb, addr, addr, SECTION_RW_NA | SECTION_WR_THROUGH);
		addr += SECTION_SIZE;
	}

	/* gpu */
	// addr = GPU;
	// for(int i = 0; i < GPU_MEM; i++)
	// {
	// 	mmu_map_section(ttb, addr, addr, SECTION_RW_NA | SECTION_NOT_EXEC);
	// 	addr += SECTION_SIZE;
	// }

	/* periféricos */
	addr = 0x20000000;
	for(int i = 0; i < 16; i++)
	{
		mmu_map_section(ttb, addr, addr, SECTION_RW_NA | SECTION_NOT_EXEC);
		addr += SECTION_SIZE;
	}

	setTranslationTableBase0(ttb);
	// debug_ttb(ttb);

	/* dominios */
	doms = getDomainAccessControl();
	doms &= ~0x0f;
	doms |= (0x01 << 2) | (0x03 << 0); // dominio nº1: cliente | dominio nº0: manager
	setDomainAccessControl(doms);

	/* mmu */
	mmu_enable();
}

uint duplicate_ttb(uint ttb)
{
	uint ttb_copy = mmap_alloc_ttb();
	for(int i = 0; i < 4096; i++) ((uint *) ttb_copy)[i] = 0;
	/* copiar kernel/periféricos */
	for(int i = 0; i < 528; i++) ((uint *) ttb_copy)[i] = ((uint *) ttb)[i];
	return ttb_copy;
}

int mmu_map_section(uint ttb, uint pAddr, uint vAddr, uint attributes)
{
	// tablas
	uint *master = (uint *) ttb;

	// índices
	int master_dx = vAddr >> 20;

	// entrada no presente
	if(master[master_dx] != FAULT)
		return -1;

	master[master_dx]  = (pAddr & ~0xfffff);
	master[master_dx] |= DOM_CLIENT << 5;
	master[master_dx] |= attributes;
	master[master_dx] |= SECTION;

	return 0;
}

int mmu_map_smallpage(uint ttb, uint pAddr, uint vAddr, uint attributes)
{
	// tablas
	uint *master = (uint *) ttb;
	uint *coarse;

	// índices
	int master_dx = vAddr >> 20;
	int coarse_dx = (vAddr & 0x000ff000) >> 12;

	// entrada no presente
	if(master[master_dx] == FAULT)
	{
		coarse = (uint *) (mmap_alloc_system() & ~0xfff);
		for(int i = 0; i < 1024; i++) coarse[i] = FAULT;
		master[master_dx]  = (uint) coarse;
		master[master_dx] |= DOM_CLIENT << 5;
		master[master_dx] |= COARSE_TABLE;
	}

	coarse = (uint *) (master[master_dx] & ~0xfff);

	// dirección virtual asignada
	if(coarse[coarse_dx] != FAULT)
		return -1;

	coarse[coarse_dx]  = (pAddr & ~0xfff);
	coarse[coarse_dx] |= attributes;
	coarse[coarse_dx] |= SMALL_PAGE;

	return 0;
}

uint mmu_unmap_smallpage(uint ttb, uint vAddr)
{
	// tablas
	uint *master = (uint *) ttb;
	uint *coarse;

	// índices
	int master_dx = vAddr >> 20;
	int coarse_dx = (vAddr & 0x000ff000) >> 12;

	uint aux;

	// entrada no presente
	if(master[master_dx] == FAULT)
		return -1;

	coarse = (uint *) (master[master_dx] & ~0xfff);
	aux = coarse[coarse_dx];
	coarse[coarse_dx] = FAULT;

	tlb_line_flush(vAddr);

	return aux;
}

uint mmu_virt2phys(uint ttb, uint virt_addr)
{
	// tablas
	uint *master = (uint *) ttb;
	uint *coarse;

	// índices
	int master_dx = virt_addr >> 20;
	int coarse_dx = (virt_addr & 0x000ff000) >> 12;

	// entrada no presente
	if(master[master_dx] == FAULT)
		return FAULT;

	coarse = (uint *) (master[master_dx] & ~0xfff);

	return (coarse[coarse_dx] & ~0xfff);
}

void mmu_delete_ttb(uint ttb)
{
	uint frame;

	for(int i = 528; i < 4096; i++)
	{
		frame = ((uint *) ttb)[i];

		if(frame != FAULT)
			mmap_free(frame);
	}

	mmap_free(ttb);
	mmap_free(ttb+SMALL_PAGE_SIZE);
	mmap_free(ttb+SMALL_PAGE_SIZE);
	mmap_free(ttb+SMALL_PAGE_SIZE);
}

/* Debug */

// TODO: Añadir impresión de atributos

void debug_section(uint section, uint *coarse)
{
	for(int i = 0; i < 256; i++)
		if(coarse[i] != 0)
		{
			debug("%x - %x | %b\n", (i << 12) + section, coarse[i] & 0xfffff000, coarse[i] & 0xfff);
		}
}

void debug_ttb(uint ttb)
{
	debug("Virtual - Physical\n");
	for(int i = 0; i < 4096; i++)
		if(((uint *) ttb)[i] != FAULT)
		{
			if(((uint *) ttb)[i] & COARSE_TABLE)
				debug_section((i << 20), (uint *) (((uint *) ttb)[i] & 0xfffff000));
			else
			{
				debug("%x - %x | %b\n", (i << 20), ((uint *) ttb)[i] & 0xfff00000, ((uint *) ttb)[i] & 0xfff);
			}
		}
}
