#include <process.h>

int last_pid = 0;

// Crea struct Process, la incializa
process_t* process_create()
{
	process_t *process = kmalloc(sizeof(process_t));

	process->pid = last_pid++;
	process->parent = 0;
	process->children = queueCreate();

	process->argv = 0;
	process->argc = 0;

	// Process Control Block
	process->pcb = kmalloc(sizeof(pcb_t));
	memoryClear(process->pcb, sizeof(pcb_t));

	// Scheduler
	process->quantum = PROC_QUANTUM;
	process->ticks = PROC_QUANTUM;
	process->state = PSTATE_RDY;

	// Segmentos de programa
	process->text = 0;
	process->rodata = 0;
	process->data = 0;
	process->bss = 0;

	// Segmentos de programa creados por el sistema
	process->svc_stack = 0; 
	process->usr_stack = 0;
	process->heap = 0;

	process->mutexes = queueCreate();
	process->shared = 0;

	return process;
}

// Copia páginas de memoria física.
void physical_frame_copy(uint dst, uint src)
{
	uint ttb;
	ttb = getTranslationTableBase0();
	mmu_map_smallpage(ttb, dst, AUX_FRAME, PAGE_RW_NA | PAGE_NOT_EXEC);
	for(int i = 0; i < 1024; i++) ((uint *) AUX_FRAME)[i] = ((uint *) src)[i];
	mmu_unmap_smallpage(ttb, AUX_FRAME);
}

// Crea el proceso init
process_t* process_init()
{
	process_t *init = process_create();

	Elf32_load(init, __init);
	
	init->svc_stack = kmalloc(sizeof(segment_t));
	init->usr_stack = kmalloc(sizeof(segment_t));

	*(init->svc_stack) = (segment_t) {
		.virt_addr = PROCESS_SVC_STACK,
		.phys_addr = 0,
		.offset = 4096,
		.attr = PAGE_RW_RW | PAGE_NOT_EXEC | PAGE_WR_THROUGH
	};

	init->usr_stack = kmalloc(sizeof(segment_t));

	*(init->usr_stack) = (segment_t) {
		.virt_addr = PROCESS_USR_STACK,
		.phys_addr = 0,
		.offset = 4096,
		.attr = PAGE_RW_RW | PAGE_NOT_EXEC | PAGE_WR_THROUGH
	};

	init->pcb->r13_usr = init->usr_stack->virt_addr + SMALL_PAGE_SIZE;
	init->pcb->r13_svc = init->svc_stack->virt_addr + SMALL_PAGE_SIZE;
	init->pcb->r14_svc = init->text->virt_addr;
	init->pcb->spsr_svc = 0x150; // PSR: Usr32md | IRQ | no FIQ | ARM state
	init->pcb->ttb = duplicate_ttb(getTranslationTableBase0());
        
	// Carga sección .text ////
	uint frame;
	uint virt_addr = init->text->virt_addr;
	uint phys_addr = init->text->phys_addr;
	uint frames = (init->text->offset >> 12) + (((init->text->offset & 0xfff) > 0)? 1 : 0);

	while(frames--)
	{
		frame = mmap_alloc_user();

		physical_frame_copy(frame, phys_addr);

		mmu_map_smallpage(init->pcb->ttb, frame, virt_addr, init->text->attr);

		virt_addr += SMALL_PAGE_SIZE;
		phys_addr += SMALL_PAGE_SIZE;
	}
	////////////////////

	mmu_map_smallpage(init->pcb->ttb, mmap_alloc_user(), init->svc_stack->virt_addr, init->svc_stack->attr);
	mmu_map_smallpage(init->pcb->ttb, mmap_alloc_user(), init->usr_stack->virt_addr, init->usr_stack->attr);
        
    heapInit(init->pcb->ttb, PROCESS_HEAP);

	//debug("%% process(%d) %%\n", init->pid);
	//debug_ttb(init->pcb->ttb);
    //debug_pcb(init->pcb);

	return init;
}

/* Elf32 - ARM */

// Identifica los segmentos de programa (funcionalidad mínima)
int Elf32_load(process_t *process, void *file)
{
	// TODO: Verificar que sea ELF32, ARM, etc

	Elf32_Ehdr *h = file;
	Elf32_Shdr *sh = (Elf32_Shdr *) ((uint) file + (uint) h->e_shoff);
	Elf32_Shdr *shstrtab = (Elf32_Shdr *) ((uint) file + (uint) h->e_shoff + (uint) h->e_shstrndx * (uint) h->e_shentsize);
	Elf32_Shdr *cursor;
	char *strtab = (char *) ((uint) file + shstrtab->sh_offset);

	for(int i = 0; i < ((uint) h->e_shnum); i++)
	{
		cursor = (Elf32_Shdr *) ((uint) sh + (uint) i * (uint) h->e_shentsize);
		char *sh_name = &(strtab[cursor->sh_name]);

		if(!stringEqual(sh_name, ((char *) (&".text")), stringLength(sh_name)))
		{
			// debug(".text virt: %x phys: %x\n", cursor->sh_offset + (uint) file, cursor->sh_addr);

			segment_t *section = kmalloc(sizeof(segment_t));

			section->virt_addr = cursor->sh_addr;
			section->phys_addr = cursor->sh_offset + (uint) file;
			section->offset = cursor->sh_size;
			section->attr = PAGE_RW_RO | PAGE_WR_THROUGH;

			process->text = section;
		}
		else if(!stringEqual(sh_name, ((char *) (&".rodata")), stringLength(sh_name)))
		{
			// debug(".rodata virt: %x phys: %x\n", cursor->sh_offset + (uint) file, cursor->sh_addr);

			segment_t *section = kmalloc(sizeof(segment_t));

			section->virt_addr = cursor->sh_addr;
			section->phys_addr = cursor->sh_offset + (uint) file;
			section->offset = cursor->sh_size;
			section->attr = PAGE_RW_RO | PAGE_NOT_EXEC | PAGE_WR_THROUGH;

			process->rodata = section;
		}
		else if(!stringEqual(sh_name, ((char *) (&".data")), stringLength(sh_name)))
		{
			// debug(".data virt: %x phys: %x\n", cursor->sh_offset + (uint) file, cursor->sh_addr);
			
			segment_t *section = kmalloc(sizeof(segment_t));

			section->virt_addr = cursor->sh_addr;
			section->phys_addr = cursor->sh_offset + (uint) file;
			section->offset = cursor->sh_size;
			section->attr = PAGE_RW_RW | PAGE_NOT_EXEC | PAGE_WR_THROUGH;

			process->data = section;
		}
		else if(!stringEqual(sh_name, ((char *) (&".bss")), stringLength(sh_name)))
		{
			// debug(".bss virt: %x phys: %x\n", cursor->sh_offset + (uint) file, cursor->sh_addr);

			segment_t *section = kmalloc(sizeof(segment_t));

			section->virt_addr = cursor->sh_addr;
			section->phys_addr = cursor->sh_offset + (uint) file;
			section->offset = cursor->sh_size;
			section->attr = PAGE_RW_RW | PAGE_NOT_EXEC | PAGE_WR_THROUGH;

			process->bss = section;
		}
	}

	return 0;
}

/* debug */

void debug_pcb(pcb_t *pcb)
{
	debug("=========begin=========\n");

	if(pcb != 0)
	{
		debug("\tr0_usr   %x\n", pcb->r0_usr);
		debug("\tr1_usr   %x\n", pcb->r1_usr);
		debug("\tr2_usr   %x\n", pcb->r2_usr);
		debug("\tr3_usr   %x\n", pcb->r3_usr);
		debug("\tr4_usr   %x\n", pcb->r4_usr);
		debug("\tr5_usr   %x\n", pcb->r5_usr);
		debug("\tr6_usr   %x\n", pcb->r6_usr);
		debug("\tr7_usr   %x\n", pcb->r7_usr);
		debug("\tr8_usr   %x\n", pcb->r8_usr);
		debug("\tr9_usr   %x\n", pcb->r9_usr);
		debug("\tr10_usr  %x\n", pcb->r10_usr);
		debug("\tr11_usr  %x\n", pcb->r11_usr);
		debug("\tr12_usr  %x\n", pcb->r12_usr);
		debug("\tr13_usr  %x\n", pcb->r13_usr);
		debug("\tr14_usr  %x\n", pcb->r14_usr);
		debug("\tr13_svc  %x\n", pcb->r13_svc);
		debug("\tr14_svc  %x\n", pcb->r14_svc);
		debug("\tspsr_svc %x\n", pcb->spsr_svc);
		debug("\tttb      %x\n", pcb->ttb);
	}

	debug("==========end==========\n");
}
